#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Function file for regression maps.

Created on Sun Sep 20 15:39:13 2020

@author: andy
"""

import numpy as np
#import xarray as xr
import matplotlib.pyplot as plt
from sklearn.linear_model import LogisticRegression

def elr_fit(x, y):
    """
    Fits an Extended Logistic Regression to a set of tercile-category hindcasts at a single point.
    Uses sklearn.linear_model.LogisticRegression
    
    Parameters
    ----------
    x : np array containing the set of GCM ensemble-mean hindcasts (rows) at a single gridpoint (1 column)
    y: np array containing the corresponding validating observations 
    
    Returns  
    A tuple containing: 
    elrmodel: class object sklearn.linear_model.LogisticRegression.fit()
    (p33, p67): tuple of the tercile breaks
    """
# NB: the training in univariate; we have a different model at each gridpoint in the current implementation;
# once could imagine a bayesian spatial hierchical model, but that's not what's done here.
# Thus, the function will be called within a loop over gripoints.
# There could also be an XV loop.

    # Tercile breaks
    p33 = np.percentile(y,100/3); p67 = np.percentile(y,100*2/3)
    # stack y_bn & y_bn_nn vecors below each other (reshaping first into col vectors)
    yc = np.vstack((np.where(y <= p33, 1, 0).reshape(-1, 1),np.where(y <= p67, 1, 0).reshape(-1, 1)))

    # Construct 2D predictor matrix, with x in col 1 and q in col 2
    ndat = np.size(x)
    qq = np.vstack((np.full(ndat, p33).reshape(-1, 1), np.full(ndat, p67).reshape(-1, 1)))
    xmat = np.hstack((np.vstack((x.reshape(-1, 1), x.reshape(-1, 1))), qq))

# Train the regression
    elr_model = LogisticRegression(fit_intercept=True).fit(xmat, np.ravel(yc))
    terciles = (p33, p67)
    return elr_model, terciles

# def ELR[L,M] = trainELR(x, y) :
#     """
#     Front end to elr_fit.
#     Train ELR model by calling elr_fit at every gridpoint of field.
#     The model is trained only at gridpints where the mean of y is non-missing (ie over land)
    
#     Parameters
#     ----------
#     x : xarray dataset containing single predictor field timeseries
#     y : xarray dataset containing predictand field timeseries
#     x & y grids must match! (add a check!)
     
#     Returns
#     -------
#      elr_model: class object sklearn.linear_model.LogisticRegression.fit()
#     """    
#     # Make a dictionary to contain the 'LogisticRegression' objects and terciles
#     elr_dict = {}   # create an empty dictionary
    
#     ym = np.mean(y,axis=0)
#     msk = ma.getmask(ym)
#     index_land = np.empty((nlat,nlon),dtype=int)
    
    
#     x0 = np.zeros(np.shape(xm))     # array of zeros to construct the climo forecast
#     ijland = -1
#     for j in range(nlat):
#     #    print("in j loop, j=", j)
#         for i in range(nlon):
#             if msk[j,i] == False:    # fit model just for landpoints
#                 ijland = ijland + 1
#                 index_land[j,i] = ijland     # index of land points
#                 elr_dict[ijland] = elr_fit(xm[:,j,i], y[:,j,i])  
#                 elr_climo_dict[ijland] = elr_fit(x0[:,j,i], y[:,j,i])  
#                 # ijland is the dictionary key that can be used to assess the entries, like this
#                 # mymodel, mytercs = mydict[0]
#                 # mymodel.coef_
#     nland = ijland+1
#     print('ELR training done with total landpoints = ',nland)



def elr_tercilesPredict(x, terciles, elr_model):
    """
    Make a set of tercile predictions from a predictor series and an extended logistic regression model.
    The tercile breaks from the training period are required.
    Uses sklearn.linear_model.LogisticRegression.predict_proba()
    
    Parameters
    ----------
    x : np array containing the set of predictors at a single gridpoint
    terciles: tuple containing the tercile breaks from the training period (p33, p67)
    elr_model: class object sklearn.linear_model.LogisticRegression.fit()
    
    Returns
    -------
    fterciles: np array containing the tercile-category forecasts (3 cols: BN, NN, AN)
    """
    ndat = np.size(x)
    f33 = elr_model.predict_proba(np.hstack((x.reshape(-1, 1), np.full(ndat, terciles[0]).reshape(-1, 1))))
    f67 = elr_model.predict_proba(np.hstack((x.reshape(-1, 1), np.full(ndat, terciles[1]).reshape(-1, 1))))
    fterciles = np.vstack( ( f33[:,1], f67[:,1]-f33[:,1], np.ones((ndat,))-f67[:,1] ) ).T
    return fterciles

def elr_poe(x, elr_model_forecast, elr_model_climo):
    """
    Probability of Exceedance Curve for Extended Logistic Regression forecast at a single gridpoint.
    Calculates PoE (poe, ordinate; 0 to 1) vs precip amounts (qclim & qfcst, abscissa).
    Uses sklearn.linear_model.LogisticRegression
    [***Under Construction***]
    
    Parameters
    ----------
    x : float containing the GCM ensemble-mean forecast value at a single gridpoint (1 column)
    elr_model_forecast: class object sklearn.linear_model.LogisticRegression containing the ELR forecast model
    elr_model_climo: As elr_model_forecast, but for a model trained with x=0 (ie the climatological forecast)
    
    Returns
    -------
    poe: array of exceedance probabilities on the domain ]0,1[
    q_fcst: array of forecast quantile values for each poe value
    q_clim: array of climatological quantile values for each poe value
    """
    
    # Forecast model parameters
    c = elr_model_forecast.intercept_
    b = elr_model_forecast.coef_
   # Climo model parameters
    C = elr_model_climo.intercept_
    B = elr_model_climo.coef_

    # Create a vector of npoints p in (0, 1) 
    npoints = 500
    p = np.linspace(0.001, 1-0.001, npoints)
    poe = 1 - p
    
    # Forecast and climo curves
    q_clim = (1/B[0,1]) * np.log( p / ((1-p) * np.exp(C)))
    q_fcst = (1/b[0,1]) * np.log( p / ((1-p) * np.exp(c + b[0,0]*x)))
    return poe, q_fcst, q_clim


def RPSS(f, po):
    """
    Calculates Ranked Probability Skill Score for tercile-category forecasts.
    
    Parameters
    ----------
    f : np array containing the set of forecast probabilities (rows) for the BN, NN, and AN categories (3 cols)
    po: np array containing the corresponding validating observations (ones and zeros) (3 cols)
    
    Returns
    -------
    RPSS score averaged across all samples
    """
    
    f1 = np.cumsum(f,axis=1) # f(sample,cat) the 2nd axis (1) needs to be the fcst category; the first (0) is the sample
    po1 = np.cumsum(po,axis=1)

    rps_for = (po1[:,0]-f1[:,0])**2 + (po1[:,1]-f1[:,1])**2 + (po1[:,2]-f1[:,2])**2
    rps_clim = (po1[:,0]-0.33)**2 + (po1[:,1]-0.67)**2 + (po1[:,2]-1.0)**2

    rpss = 1 - np.nanmean(rps_for)/np.nanmean(rps_clim)
    return rpss

def myRPSSmapPlot(feld, lons, lats, figtitle, fname):
    """
    Map plotting
    
    Parameters
    ----------
    feld : np array containing the 2D field to plot
    lons: np 1D array of longitudes
    lats: np 1D array of latitudes
    figtitle: title string
    fname: output file name, including full path and .pdf extension
    
    Returns
    -------
    The plot is saved to the given filename
    """
    import cartopy.crs as ccrs
    from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
    import cartopy.feature as cfeature
    
    fig, ax = plt.subplots()
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.coastlines(resolution='50m')
    ax.add_feature(cfeature.BORDERS, linewidth=0.5)

    lons2, lats2 = np.meshgrid(lons, lats)
    CS = ax.contourf(lons2, lats2, feld, 10, cmap=None, extend='both',
                     levels=[-10, -5, -1.25, 1.25, 2.5, 5, 7.5, 10, 12.5, 15, 17.5, 20],
                     colors = [(0.11, 0.57, 1.0),(0.0,1.0,1.0),(0.75, 0.75, 0.75),
                               (1.0, 1.0, 0.0),(1.0, 0.84, 0.0),(1.0, 0.65, 0.0),
                               (1.0, 0.55, 0.0),(1.0, 0.39, 0.28),(1.0, 0.27, 0.0),
                               (1.0, 0.0, 0.0),(1.0, 0.0, 1.0)] )
#    CS.cmap.set_under((0.11, 0.57, 1.0))
#    CS.cmap.set_over((1.0, 0.0, 1.0))
    cbar = plt.colorbar(CS, shrink=0.7, orientation='vertical')
    cbar.set_ticks([ -5, 0, 5, 10, 15, 20],)
    cbar.set_ticklabels(["-5", "0", "5", "10", "15", "20"])
    #CS = ax.contour(lons2, lats2, rpss_ELR*100, 10, linewidths=0.5, colors='k')
    #ax.clabel(CS, inline=1, fontsize=8, fmt='%2.0f')

    ax.set_title(figtitle)
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, 
#                      xlocs=[-20, -15, -10, -5, 0, 5, 10], ylocs=[0, 5, 10, 15, 20],
                      linewidth=1, color='gray', alpha=0.5, linestyle='--')
    gl.xlabels_top = False
    gl.ylabels_right = False
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER

    plt.savefig(fname)
    return

